#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

import torch
import shutil
import pytest
from transformers import AutoTokenizer
from lmformatenforcer import JsonSchemaParser
from lmformatenforcer.integrations.transformers import build_token_enforcer_tokenizer_data

from happy_vllm import utils
from happy_vllm.logits_processors import json_format as json_format_module
from happy_vllm.logits_processors.json_format import VLLMLogitsProcessorJSON

from .conftest import TEST_MODELS_DIR


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def test_check_simple_json():
    # Nominal case
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['number']}
    assert json_format_module.check_simple_json(simple_json)
    
    # One wrong value
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'float',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['number']}
    with pytest.raises(ValueError) as error:
        json_format_module.check_simple_json(simple_json)

    # One wrong value in a list
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['float']}
    with pytest.raises(ValueError) as error:
        json_format_module.check_simple_json(simple_json)
    
    # List with zero element
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': []}
    with pytest.raises(ValueError) as error:
        json_format_module.check_simple_json(simple_json)

    # List with more than one element
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['string', 'integer']}
    with pytest.raises(ValueError) as error:
        json_format_module.check_simple_json(simple_json)

    # One wrong keyword
    simple_json = {1: 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number'}
    with pytest.raises(ValueError) as error:
        json_format_module.check_simple_json(simple_json)


def test_get_json_format_parser():
    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['number']}
    json_schema = {"properties": 
               {"name": {"type": "string"},
                "surname": {"type": "string"},
                "favorite_food": {"type": "string"},
                "current_number_of_children": {"type": "integer"},
                "work": {"type": "object", "properties": {"team_name": {"type": "string"}, "job_title": {"type": "string"}}, "required":["team_name", "job_title"]}},
                "required": ["name", "surname", "favorite_food", "current_number_of_children", "work"], "title": "json_format", "type": "object"}
    
    # Nominal case, simple json
    json_format = simple_json
    json_format_parser = json_format_module.get_json_format_parser(json_format=json_format, is_json_schema=False)
    assert isinstance(json_format_parser, JsonSchemaParser)

    # Nominal case, json schema
    json_format = json_schema
    json_format_parser = json_format_module.get_json_format_parser(json_format=json_format, is_json_schema=True)
    assert isinstance(json_format_parser, JsonSchemaParser)

    # Wrong is_json_schema for a simple json
    json_format = simple_json
    with pytest.raises(Exception) as error:
        json_format_parser = json_format_module.get_json_format_parser(json_format=json_format, is_json_schema=True)

    # Wrong is_json_schema for a json schema
    json_format = json_schema
    with pytest.raises(ValueError) as error:
        json_format_parser = json_format_module.get_json_format_parser(json_format=json_format, is_json_schema=False)


def test_VLLMLogitsProcessorJSON_init():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)
    tokenizer_lmformatenforcer = build_token_enforcer_tokenizer_data(tokenizer)

    simple_json = {'first_keyword': 'string',
                    'second_keyword': 'integer',
                    'third_keyword': 'boolean',
                    'fourth_keyword': 'number',
                    'fifth_keyword': ['string'],
                    'sixth_keyword': ['integer'],
                    'seventh_keyword': ['boolean'],
                    'eighth_keyword': ['number']}
    json_schema = {"properties": 
               {"name": {"type": "string"},
                "surname": {"type": "string"},
                "favorite_food": {"type": "string"},
                "current_number_of_children": {"type": "integer"},
                "work": {"type": "object", "properties": {"team_name": {"type": "string"}, "job_title": {"type": "string"}}, "required":["team_name", "job_title"]}},
                "required": ["name", "surname", "favorite_food", "current_number_of_children", "work"], "title": "json_format", "type": "object"}

    # Nominal case simple json
    json_format = simple_json
    json_format_is_json_schema = False
    json_logits_processor = VLLMLogitsProcessorJSON(tokenizer_lmformatenforcer, json_format, json_format_is_json_schema)
    assert isinstance(json_logits_processor, VLLMLogitsProcessorJSON)
    assert json_logits_processor.logits_processor_activated

    # Nominal case json schema
    json_format = json_schema
    json_format_is_json_schema = True
    json_logits_processor = VLLMLogitsProcessorJSON(tokenizer_lmformatenforcer, json_format, json_format_is_json_schema)
    assert isinstance(json_logits_processor, VLLMLogitsProcessorJSON)
    assert json_logits_processor.logits_processor_activated

    # Nominal case empty json format
    json_format = {}
    json_format_is_json_schema = True
    json_logits_processor = VLLMLogitsProcessorJSON(tokenizer_lmformatenforcer, json_format, json_format_is_json_schema)
    assert isinstance(json_logits_processor, VLLMLogitsProcessorJSON)
    assert not(json_logits_processor.logits_processor_activated)

    # Nominal case json format is None
    json_format = None
    json_format_is_json_schema = False
    json_logits_processor = VLLMLogitsProcessorJSON(tokenizer_lmformatenforcer, json_format, json_format_is_json_schema)
    assert isinstance(json_logits_processor, VLLMLogitsProcessorJSON)
    assert not(json_logits_processor.logits_processor_activated)


def test_VLLMLogitsProcessorJSON_call():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)
    tokenizer_lmformatenforcer = build_token_enforcer_tokenizer_data(tokenizer)

    # Logits processor is deactivated
    json_format = None
    json_logits_processor = VLLMLogitsProcessorJSON(tokenizer_lmformatenforcer, json_format)
    scores = torch.rand((1000, ))
    assert max(scores) == pytest.approx(max(json_logits_processor([], scores)))
    assert min(scores) == pytest.approx(min(json_logits_processor([], scores)))


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import torch
import shutil
import pytest
import numpy as np
from transformers import AutoTokenizer

from happy_vllm import utils
from happy_vllm.logits_processors.response_pool import VLLMLogitsProcessorResponsePool

from .conftest import TEST_MODELS_DIR


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def test_VLLMLogitsProcessorResponsePool_init():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)

    possible_responses = ['Yes', 'No', "I don't know"]
    logit_processor_response_pool = VLLMLogitsProcessorResponsePool(tokenizer, possible_responses)
    for possible_response in possible_responses:
        assert possible_response in logit_processor_response_pool.possible_tokens_responses
        tokens_in_processor = logit_processor_response_pool.possible_tokens_responses[possible_response]
        direct_tokens = list(utils.proper_tokenization(tokenizer, possible_response))
        assert tokens_in_processor == direct_tokens
        assert logit_processor_response_pool.eos_token_id == tokenizer.eos_token_id


def test_VLLMLogitsProcessorResponsePool_call():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)

    possible_responses = ['Yes', 'No', 'I am not really sure', 'I am not certain']
    logit_processor_response_pool = VLLMLogitsProcessorResponsePool(tokenizer, possible_responses)
    max_token_ids = max([max(tokens_ids) for tokens_ids in logit_processor_response_pool.possible_tokens_responses.values()])
    scores = torch.tensor(range(max_token_ids+1000), dtype=float)
    
    # No input_ids yet
    new_scores = logit_processor_response_pool([], scores)
    first_tokens = {possible_responses_ids[0] for possible_responses_ids in logit_processor_response_pool.possible_tokens_responses.values()}        
    for token in first_tokens:
        assert pytest.approx(scores[token]) == new_scores[token]
    list_other_tokens = [token for token in range(scores.shape[-1]) if token not in first_tokens]
    assert max(new_scores[list_other_tokens]) == -np.inf

    # Began the response 'I am not really sure'
    new_scores = logit_processor_response_pool(logit_processor_response_pool.possible_tokens_responses['I am not really sure'][:-1], scores)
    token_to_test = logit_processor_response_pool.possible_tokens_responses['I am not really sure'][-1]
    assert pytest.approx(scores[token_to_test]) == new_scores[token_to_test]
    list_other_tokens = [token for token in range(scores.shape[-1]) if token!=token_to_test]
    assert max(new_scores[list_other_tokens]) == -np.inf

    # Began the response 'I am not' (same beginning between 'I am not really sure' and 'I am not certain')
    tokens_i_am_not_really_sure = logit_processor_response_pool.possible_tokens_responses['I am not really sure']
    tokens_i_am_not_certain = logit_processor_response_pool.possible_tokens_responses['I am not certain']
    i = 0
    while tokens_i_am_not_really_sure[i] == tokens_i_am_not_certain[i]:
        i += 1
    new_scores = logit_processor_response_pool(tokens_i_am_not_really_sure[:i], scores)
    tokens_to_test = {tokens_i_am_not_really_sure[i], tokens_i_am_not_certain[i]}
    for token in tokens_to_test:
        assert pytest.approx(scores[token]) == new_scores[token]
    list_other_tokens = [token for token in range(scores.shape[-1]) if token not in tokens_to_test]
    assert max(new_scores[list_other_tokens]) == -np.inf

    # The model generated a complete response -> eos-token
    new_scores = logit_processor_response_pool(logit_processor_response_pool.possible_tokens_responses['I am not really sure'], scores)
    eos_token_id = tokenizer.eos_token_id
    assert scores[eos_token_id] == new_scores[eos_token_id]
    list_other_tokens = [token for token in range(scores.shape[-1]) if token!=eos_token_id]
    assert max(new_scores[list_other_tokens]) == -np.inf

    # Empty list
    possible_responses = []
    logit_processor_response_pool = VLLMLogitsProcessorResponsePool(tokenizer, possible_responses)
    scores = torch.tensor(range(1000), dtype=float)
    assert max(scores) == pytest.approx(max(logit_processor_response_pool([], scores)))
    assert min(scores) == pytest.approx(min(logit_processor_response_pool([], scores)))


def test_VLLMLogitsProcessorResponsePool_get_next_possible_tokens():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)

    possible_responses = ['Yes', 'No', 'I am not really sure', 'I am not certain']
    logit_processor_response_pool = VLLMLogitsProcessorResponsePool(tokenizer, possible_responses)

    # No input_ids yet
    next_tokens = logit_processor_response_pool._get_next_possible_tokens([])
    first_tokens = {possible_responses_ids[0] for possible_responses_ids in logit_processor_response_pool.possible_tokens_responses.values()}        
    assert first_tokens == set(next_tokens)

    # Began the response 'I am not really sure'
    next_tokens = logit_processor_response_pool._get_next_possible_tokens(logit_processor_response_pool.possible_tokens_responses['I am not really sure'][:-1])
    token_to_test = logit_processor_response_pool.possible_tokens_responses['I am not really sure'][-1]       
    assert {token_to_test} == set(next_tokens)

    # Began the response 'I am not' (same beginning between 'I am not really sure' and 'I am not certain')
    tokens_i_am_not_really_sure = logit_processor_response_pool.possible_tokens_responses['I am not really sure']
    tokens_i_am_not_certain = logit_processor_response_pool.possible_tokens_responses['I am not certain']
    i = 0
    while tokens_i_am_not_really_sure[i] == tokens_i_am_not_certain[i]:
        i += 1
    next_tokens = logit_processor_response_pool._get_next_possible_tokens(tokens_i_am_not_really_sure[:i])
    tokens_to_test = {tokens_i_am_not_really_sure[i], tokens_i_am_not_certain[i]}
    assert set(next_tokens) == tokens_to_test

    # The model generated a complete response -> eos-token
    next_tokens = logit_processor_response_pool._get_next_possible_tokens(logit_processor_response_pool.possible_tokens_responses['I am not really sure'])
    assert set(next_tokens) == {tokenizer.eos_token_id}


def test_VLLMLogitsProcessorResponsePool_get_common_tokens_ids_end_begin():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)

    possible_responses = ['Yes', 'No', 'I am not really sure', 'I am not certain']
    logit_processor_response_pool = VLLMLogitsProcessorResponsePool(tokenizer, possible_responses)

    list_end = [1, 2, 3, 4]
    list_begin = [3, 4, 5, 6]
    assert 2 == logit_processor_response_pool._get_common_tokens_ids_end_begin(list_end, list_begin)

    list_end = [1, 2, 3, 4, 2]
    list_begin = [3, 4, 5, 6]
    assert 0 == logit_processor_response_pool._get_common_tokens_ids_end_begin(list_end, list_begin)

    list_end = [1, 2, 3]
    list_begin = [3, 4, 5, 6]
    assert 1 == logit_processor_response_pool._get_common_tokens_ids_end_begin(list_end, list_begin)

    list_end = [1, 2, 3]
    list_begin = [4, 5, 6]
    assert 0 == logit_processor_response_pool._get_common_tokens_ids_end_begin(list_end, list_begin)

    list_end = list(range(100))
    list_begin = list(range(100))
    assert 100 == logit_processor_response_pool._get_common_tokens_ids_end_begin(list_end, list_begin)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import pytest

from transformers import AutoTokenizer
from fastapi.testclient import TestClient
from vllm.sampling_params import SamplingParams
from lmformatenforcer.integrations.transformers import build_token_enforcer_tokenizer_data


from happy_vllm import utils
from .conftest import TEST_MODELS_DIR
from happy_vllm.routers import functional
from happy_vllm.model.model_base import Model
from happy_vllm.logits_processors.json_format import VLLMLogitsProcessorJSON
from happy_vllm.logits_processors.response_pool import VLLMLogitsProcessorResponsePool


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def init_model(truncation_side="left"):
    model = Model(app_name=os.environ['app_name'])
    model._tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR, truncation_side=truncation_side)
    model._tokenizer_lmformatenforcer = build_token_enforcer_tokenizer_data(model._tokenizer)
    model.original_truncation_side = truncation_side
    model.max_model_len = 2048
    return model


def test_parse_logits_processors():
    """Tests the function parse_logits_processors"""
    model = init_model()
    tokenizer = model._tokenizer
    tokenizer_lmformatenforcer = model._tokenizer_lmformatenforcer
    
    # With response_pool
    prompt = "This is the prompt"
    for response_pool in [["Yes", "No"], ["Certainly", "I think so", "Of course"], ["Without a doubt"]]:
        request_dict = {"temperature": 0, "keyword": "value", "response_pool": response_pool}
        logits_processors = functional.parse_logits_processors(request_dict, prompt, model, tokenizer, tokenizer_lmformatenforcer)
        assert request_dict == {"temperature": 0, "keyword": "value"}
        assert len(logits_processors) == 1
        assert isinstance(logits_processors[0], VLLMLogitsProcessorResponsePool)
        assert len(logits_processors[0].possible_tokens_responses) == len(response_pool)
        assert logits_processors[0].eos_token_id == tokenizer.eos_token_id

    # With json_format
    prompt = "This is the prompt"
    json_format = {"name": "string",
                    "surname": "string",
                    "favorite_food": "string",
                    "current_number_of_children": "integer",
                    "team_name": "string",
                    "job_title": "string"}
    request_dict = {"temperature": 0, "keyword": "value", "json_format": json_format}
    logits_processors = functional.parse_logits_processors(request_dict, prompt, model, tokenizer, tokenizer_lmformatenforcer)
    assert request_dict == {"temperature": 0, "keyword": "value"}
    assert len(logits_processors) == 1
    assert isinstance(logits_processors[0], VLLMLogitsProcessorJSON)
    assert logits_processors[0].logits_processor_activated

    # With json_format
    prompt = "This is the prompt"
    json_format = {}
    request_dict = {"temperature": 0, "keyword": "value", "json_format": json_format}
    logits_processors = functional.parse_logits_processors(request_dict, prompt, model, tokenizer, tokenizer_lmformatenforcer)
    assert request_dict == {"temperature": 0, "keyword": "value"}
    assert len(logits_processors) == 1
    assert isinstance(logits_processors[0], VLLMLogitsProcessorJSON)
    assert not(logits_processors[0].logits_processor_activated)
            
    # Without logits_processors
    request_dict = {"temperature": 0, "keyword": "value"}
    logits_processors = functional.parse_logits_processors(request_dict, prompt, model, tokenizer, tokenizer_lmformatenforcer)
    assert request_dict == {"temperature": 0, "keyword": "value"}
    assert logits_processors == []


def test_parse_generate_parameters():
    """Tests the function parse_generate_parameters"""
    model = init_model()
    tokenizer = model._tokenizer
    tokenizer_lmformatenforcer = model._tokenizer_lmformatenforcer
    
    # with response_pool
    prompt_ini = "This is the prompt"
    temperature = 0
    request_dict = {"temperature": temperature, "response_pool": ["Yes", "No"], "prompt": prompt_ini, "max_tokens": 100}
    prompt, prompt_in_response, sampling_params = functional.parse_generate_parameters(request_dict, model, tokenizer, tokenizer_lmformatenforcer)
    assert prompt == prompt_ini
    assert not(prompt_in_response)
    assert isinstance(sampling_params, SamplingParams)
    assert sampling_params.temperature == temperature
    assert len(sampling_params.logits_processors) == 1
    assert isinstance(sampling_params.logits_processors[0], VLLMLogitsProcessorResponsePool)

    # with json_format
    prompt_ini = "This is the prompt"
    temperature = 0
    json_format = {"name": "string",
                    "surname": "string",
                    "favorite_food": "string",
                    "current_number_of_children": "integer",
                    "team_name": "string",
                    "job_title": "string"}
    request_dict = {"temperature": temperature, "json_format": json_format, "prompt": prompt_ini, "max_tokens": 100}
    prompt, prompt_in_response, sampling_params = functional.parse_generate_parameters(request_dict, model, tokenizer, tokenizer_lmformatenforcer)
    assert prompt == prompt_ini
    assert not(prompt_in_response)
    assert isinstance(sampling_params, SamplingParams)
    assert sampling_params.temperature == temperature
    assert len(sampling_params.logits_processors) == 1
    assert isinstance(sampling_params.logits_processors[0], VLLMLogitsProcessorJSON)

    # with no logits_processors and prompt_in_response=True
    prompt_ini = "This is the prompt but it is longer than before"
    temperature = 1.5
    max_tokens = 123
    request_dict = {"temperature": temperature, "prompt": prompt_ini, "max_tokens": max_tokens, "prompt_in_response": True}
    prompt, prompt_in_response, sampling_params = functional.parse_generate_parameters(request_dict, model, tokenizer, tokenizer_lmformatenforcer)
    assert prompt == prompt_ini
    assert prompt_in_response
    assert isinstance(sampling_params, SamplingParams)
    assert sampling_params.temperature == temperature
    assert sampling_params.max_tokens == max_tokens
    assert sampling_params.logits_processors == []

    # with no logits_processors and prompt_in_response=False
    prompt_ini = "Here"
    temperature = 0.1
    max_tokens = 1234
    request_dict = {"temperature": temperature, "prompt": prompt_ini, "max_tokens": max_tokens, "prompt_in_response": False}
    prompt, prompt_in_response, sampling_params = functional.parse_generate_parameters(request_dict, model, tokenizer, tokenizer_lmformatenforcer)
    assert prompt == prompt_ini
    assert not(prompt_in_response)
    assert isinstance(sampling_params, SamplingParams)
    assert sampling_params.temperature == temperature
    assert sampling_params.max_tokens == max_tokens
    assert sampling_params.logits_processors == []

    # Raise ValueError if response_pool and min_tokens are present
    prompt_ini = "This is the prompt"
    temperature = 0
    request_dict = {"temperature": temperature, "response_pool": ["Yes", "No"], "prompt": prompt_ini, "min_tokens": 10, "max_tokens": 100}
    with pytest.raises(ValueError):
        prompt, prompt_in_response, sampling_params = functional.parse_generate_parameters(request_dict, model, tokenizer, tokenizer_lmformatenforcer)


def test_generate(test_complete_client: TestClient):
    """Test the route generate thanks to the test_complete_client we created in conftest.py"""
    model = init_model()
    tokenizer = model._tokenizer

    def get_response(tokenizer, prompt, i, max_tokens):
        prompt_tot = f"n={i} "*i + prompt + " This is the generated text. I find it really good don't you ?"
        token_ids = tokenizer(prompt_tot, add_special_tokens=False)['input_ids']
        token_ids = token_ids[:max_tokens]
        return tokenizer.decode(token_ids)
    
    # Nominal case
    max_tokens = 500
    prompt = "Hey"
    body = {"prompt": prompt, "max_tokens": max_tokens}
    response = test_complete_client.post("/tests/generate", json=body)
    assert response.status_code == 200
    response_json = response.json()
    assert response_json["responses"] == [get_response(tokenizer, prompt, 0, max_tokens)]
    assert response_json["finish_reasons"] == ["stop"]
    assert set(response_json) == {"responses", "finish_reasons"}

    # Several responses and prompt_in_response
    max_tokens = 500
    prompt = "Hello there"
    body = {"prompt": prompt, "max_tokens": max_tokens, "n": 3, "prompt_in_response": True}
    response = test_complete_client.post("/tests/generate", json=body)
    assert response.status_code == 200
    response_json = response.json()
    assert len(response_json["responses"]) == 3
    assert len(response_json["finish_reasons"]) == 3
    for i in [0, 1, 2]:
        target_response = get_response(tokenizer, prompt, i, max_tokens)
        assert response_json["responses"][i] == target_response
        if target_response[-4:] == "ou ?":
            assert response_json["finish_reasons"][i] == "stop"
        else:
            assert response_json["finish_reasons"][i] == "length"
    assert response_json["prompt"] == "Hello there"
    assert set(response_json) == {"responses", "finish_reasons", "prompt"}

    # Generations stopped
    max_tokens = 5
    prompt = "Hey"
    body = {"prompt": prompt, "max_tokens": max_tokens, "n": 3}
    response = test_complete_client.post("/tests/generate", json=body)
    assert response.status_code == 200
    response_json = response.json()
    assert len(response_json["responses"]) == 3
    assert len(response_json["finish_reasons"]) == 3
    for i in [0, 1, 2]:
        target_response = get_response(tokenizer, prompt, i, max_tokens)
        assert response_json["responses"][i] == target_response
        if target_response[-4:] == "ou ?":
            assert response_json["finish_reasons"][i] == "stop"
        else:
            assert response_json["finish_reasons"][i] == "length"
    assert set(response_json) == {"responses", "finish_reasons"}

    # Some Generation stopped
    max_tokens = 18
    prompt = "Hey"
    body = {"prompt": prompt, "max_tokens": max_tokens, "n": 3}
    response = test_complete_client.post("/tests/generate", json=body)
    assert response.status_code == 200
    response_json = response.json()
    assert len(response_json["responses"]) == 3
    assert len(response_json["finish_reasons"]) == 3
    for i in [0, 1, 2]:
        target_response = get_response(tokenizer, prompt, i, max_tokens)
        assert response_json["responses"][i] == target_response
        if target_response[-4:] == "ou ?":
            assert response_json["finish_reasons"][i] == "stop"
        else:
            assert response_json["finish_reasons"][i] == "length"
    assert set(response_json) == {"responses", "finish_reasons"}


def test_tokenizer(test_complete_client: TestClient):
    """Test the functional route /tokenizer"""
    model = init_model()
    # Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        body = {"text": text, "vanilla": True}
        response = test_complete_client.post("/tests/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(text)['input_ids']
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        assert set(response_json) == {"tokens_ids", "tokens_nb"}

        # With with_tokens_str
        body = {"text": text, "with_tokens_str": True, "vanilla": True}
        response = test_complete_client.post("/tests/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = model._tokenizer(text)['input_ids']
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        tokens_str = model._tokenizer.convert_ids_to_tokens(target_tokens_ids)
        assert response_json["tokens_str"] == tokens_str
        assert set(response_json) == {"tokens_ids", "tokens_nb", "tokens_str"}

    #Non Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        body = {"text": text, "vanilla": False}
        response = test_complete_client.post("/tests/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = list(utils.proper_tokenization(model._tokenizer, text))
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        assert set(response_json) == {"tokens_ids", "tokens_nb"}

        # With with_tokens_str
        body = {"text": text, "with_tokens_str": True, "vanilla": False}
        response = test_complete_client.post("/tests/tokenizer", json=body)
        assert response.status_code == 200
        response_json = response.json()
        target_tokens_ids = list(utils.proper_tokenization(model._tokenizer, text))
        assert response_json["tokens_ids"] == target_tokens_ids
        assert response_json["tokens_nb"] == len(target_tokens_ids)
        tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in target_tokens_ids]
        assert response_json["tokens_str"] == tokens_str
        assert set(response_json) == {"tokens_ids", "tokens_nb", "tokens_str"}


def test_decode(test_complete_client: TestClient):
    """Test the functional route /decode"""
    model = init_model()

    # Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        token_ids = model._tokenizer(text)['input_ids']

        body = {'token_ids': token_ids, "vanilla": True}
        response = test_complete_client.post("/tests/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == model._tokenizer.decode(token_ids)
        assert set(response_json) == {"decoded_string"}

        # With with_tokens_str
        body = {'token_ids': token_ids, "vanilla": True, "with_tokens_str": True}
        response = test_complete_client.post("/tests/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == model._tokenizer.decode(token_ids)
        assert response_json["tokens_str"] == model._tokenizer.convert_ids_to_tokens(token_ids)
        assert set(response_json) == {"decoded_string", "tokens_str"}

    #Non Vanilla
    for text in ["How do you do?", "I am Lliam. How are you ?", "Marvelous, it works !"]:
        token_ids = model._tokenizer(text)['input_ids']

        body = {'token_ids': token_ids, "vanilla": False}
        response = test_complete_client.post("/tests/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == utils.proper_decode(model._tokenizer, token_ids)
        assert set(response_json) == {"decoded_string"}

        # With with_tokens_str
        body = {'token_ids': token_ids, "vanilla": False, "with_tokens_str": True}
        response = test_complete_client.post("/tests/decode", json=body)
        assert response.status_code == 200
        response_json = response.json()
        assert response_json["decoded_string"] == utils.proper_decode(model._tokenizer, token_ids)
        assert response_json["tokens_str"] == [utils.proper_decode(model._tokenizer, token_id) for token_id in token_ids]
        assert set(response_json) == {"decoded_string", "tokens_str"}


def test_split_text(test_complete_client: TestClient):
    """Test the route split_text thanks to the test_complete_client we created in conftest.py"""
    text = "Hey, my name is LLM. How are you ? Fine, you ? That's wonderful news : I'm also fine. But do you think it will last ?"

    body = {"text": text,
            "num_tokens_in_chunk": 2}
    response = test_complete_client.post("/tests/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ?", 
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 6}
    response = test_complete_client.post("/tests/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = ["Hey, my name is LLM.",
                        " How are you ? Fine, you ?",
                        " That's wonderful news : I'm also fine.",
                        " But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 1000}
    response = test_complete_client.post("/tests/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = [text]
    assert json_response == {"split_text": target_split_text}

    body = {"text": text,
            "num_tokens_in_chunk": 2,
            "separators": [" ?"]}
    response = test_complete_client.post("/tests/split_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    target_split_text = target_split_text = ["Hey, my name is LLM. How are you ?",
                        " Fine, you ?",
                        " That's wonderful news : I'm also fine. But do you think it will last ?"]
    assert json_response == {"split_text": target_split_text}


def test_metadata_text(test_complete_client: TestClient):
    """Test the route metadata_text thanks to the test_complete_client we created in conftest.py"""
    text = "Hey, my name is LLM. How are you ? Fine, and you ? Great."
    model = init_model()
    tokenizer = model._tokenizer

    text_tot = text * 73
    truncation_side = "left"
    max_length = 1234
    body = {"text": text_tot,
            "truncation_side": truncation_side,
            "max_length": max_length}
    response = test_complete_client.post("/tests/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text

    text_tot = text * 73
    truncation_side = "right"
    max_length = 1234
    body = {"text": text_tot,
            "truncation_side": truncation_side,
            "max_length":1234}
    response = test_complete_client.post("/tests/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text

    text_tot = text
    truncation_side = "left"
    max_length = 1234
    body = {"text": text,
            "truncation_side": "left",
            "max_length":1234}
    response = test_complete_client.post("/tests/metadata_text", json=body)
    assert response.status_code == 200
    json_response = response.json()
    truncated_text = model.extract_text_outside_truncation(text_tot, truncation_side, max_length)
    assert json_response["tokens_nb"] == len(utils.proper_tokenization(tokenizer, text_tot))
    assert json_response["truncated_text"] == truncated_text


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import importlib.metadata
from transformers import AutoTokenizer

from happy_vllm import utils

from .conftest import TEST_MODELS_DIR


def teardown_module():
    if os.path.exists(TEST_MODELS_DIR):
        if os.path.isdir(TEST_MODELS_DIR):
            shutil.rmtree(TEST_MODELS_DIR)


def test_get_package_version():
    # Nominal case
    version = utils.get_package_version()
    assert version == importlib.metadata.version("happy_vllm")


def test_get_vllm_version():
    # Nominal case
    version = utils.get_vllm_version()
    assert version == importlib.metadata.version("vllm")


def test_proper_tokenization():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)
    for string in ['.', '!', '?', 'Hey', 'hey']:
        assert utils.proper_tokenization(tokenizer, f' {string}') != utils.proper_tokenization(tokenizer, f'{string}')
    for string in ['.', '!', '?']:
        assert utils.proper_tokenization(tokenizer, f'reallybigword{string}')[-1] == utils.proper_tokenization(tokenizer, f'{string}')[-1]
        assert utils.proper_tokenization(tokenizer, f'reallybigword {string}')[-1] == utils.proper_tokenization(tokenizer, f' {string}')[-1]

def test_proper_decode():
    tokenizer = AutoTokenizer.from_pretrained(os.environ["tokenizer_name"],
                                                     cache_dir=TEST_MODELS_DIR)
    example = "This is an example. It is very important"
    for i in range(len(example)):
        truncated_example = example[i:]
        token_ids = list(utils.proper_tokenization(tokenizer, truncated_example))
        assert truncated_example == utils.proper_decode(tokenizer, token_ids)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Annotated
from fastapi import APIRouter
from pydantic import BaseModel, Field
from starlette.responses import JSONResponse

from ..core.resources import RESOURCES, RESOURCE_MODEL
from ..model.model_base import Model
from .schemas.technical import ResponseInformation, ResponseLiveness, ResponseReadiness, ResponseLiveMetrics

from happy_vllm import utils

# Technical router
router = APIRouter()


@router.get(
    "/liveness",
    response_model=ResponseLiveness,
    name="liveness",
    tags=["technical"],
)
async def get_liveness() -> ResponseLiveness:
    """Liveness probe for k8s"""
    liveness_msg = ResponseLiveness(alive="ok")
    return liveness_msg


@router.get(
    "/readiness",
    response_model=ResponseReadiness,
    name="readiness",
    tags=["technical"],
)
async def get_readiness() -> ResponseReadiness:
    """Readiness probe for k8s"""
    model: Model = RESOURCES.get(RESOURCE_MODEL)

    if model and model.is_model_loaded():
        return ResponseReadiness(ready="ok")
    else:
        return ResponseReadiness(ready="ko")


@router.get(
    "/info",
    response_model=ResponseInformation,
    name="information",
    tags=["technical"],
)
async def info() -> ResponseInformation:
    """Rest resource for info"""
    model: Model = RESOURCES.get(RESOURCE_MODEL)

    return ResponseInformation(
        application=model.app_name,
        version=utils.get_package_version(),
        model_name=model._model_conf.get("model_name", "?"),
        vllm_version=utils.get_vllm_version(),
        truncation_side=model.original_truncation_side,
        max_length=model.max_model_len
    )


@router.get("/live_metrics", response_model=ResponseLiveMetrics)
async def get_live_metrics() -> JSONResponse:
    """Deprecated. You should use now the /metrics endpoint
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)

    gpu_cache_usage = model.get_gpu_kv_cache_usage()
    cpu_cache_usage = model.get_cpu_kv_cache_usage()
    metrics = model.get_status_requests()
    metrics["gpu_cache_usage"] = gpu_cache_usage
    metrics["cpu_cache_usage"] = cpu_cache_usage
    return JSONResponse(metrics)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math
import torch
from typing import List

from happy_vllm import utils


class VLLMLogitsProcessorResponsePool:

    def __init__(self, tokenizer, possible_responses: List[str]) -> None:
        '''Initializes the logits processor with the token ids of the possible responses

        Args:
            tokenizer : The tokenizer used by the model
            possible_responses (list) : The list containing all the possible responses from which the llm will choose from

        '''
        self.possible_tokens_responses = {response: list(utils.proper_tokenization(tokenizer, response)) for response in possible_responses}
        self.eos_token_id = tokenizer.eos_token_id

    def __call__(self, input_ids: List[int], scores: torch.Tensor) -> torch.Tensor:
        """Puts a mask to allow only tokens which could provide a response in the pool

        Args:
            input_ids (list) : The list of tokens ids generated by the LLM
            scores (torch.tensor) : The tensor containing the logits for the next token
        
        Returns:
            toch.tensor : The updated scores ie the initial scores masked to only allow possible tokens

        """
        if self.possible_tokens_responses == {}:
            return scores
        # Gets the next possible tokens
        allowed_tokens = self._get_next_possible_tokens(input_ids)
        # Compute the corresponding mask
        mask = torch.full_like(scores, -math.inf)
        mask[allowed_tokens] = 0
        # Mask the scores
        scores = scores + mask
        return scores

    def _get_next_possible_tokens(self, input_ids: List[int]) -> List[int]:
        """Gets the next possible tokens for the response to be in the allowed pool

        Args:
            inputs_ids (list) : The list of tokens ids generated by the LLM
        
        Returns:
            list : The list of possible following tokens
        """
        allowed_tokens = set()
        # If we are not "in" a response (ie this is the first token generated)
        if len(input_ids) == 0:
            # We allow all first tokens of all possible responses
            allowed_tokens = {possible_responses_ids[0] for possible_responses_ids in self.possible_tokens_responses.values()}
        else:
            # For each possible responses...
            for possible_responses, possible_responses_ids in self.possible_tokens_responses.items():
                # Get the index at which the input corresponds to this response
                i_current_token = self._get_common_tokens_ids_end_begin(input_ids, possible_responses_ids)
                # If not 0, then, we are "in" this response
                if i_current_token != 0:
                    # We are at the end of the respons so -> eos token
                    if i_current_token >= len(possible_responses_ids):
                        allowed_tokens.add(self.eos_token_id)
                    # We are in the middle of the response -> give the following token in the response
                    else:
                        allowed_tokens.add(possible_responses_ids[i_current_token])
        allowed_tokens = list(allowed_tokens)
        return allowed_tokens

    def _get_common_tokens_ids_end_begin(self, list_end: list, list_begin: list) -> int:
        """Gives the index where we are with common elements at the end of list_end and at the beginning of list_begin. For
        example if list_end = [1, 2, 3, 4] and list_begin = [3, 4, 5] we return 2

        Args:
            list_end (list) : The list where we look at the end
            list_begin (list) : The list where we look at the beginning
        
        Returns:
            int : The index (or number) of common elements at the end of list_end and at the beginning of list_begin

        """
        i = len(list_begin)
        while list_begin[:i] != list_end[-i:] and i > 0:
            i -= 1
        return i



#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from happy_vllm.logits_processors.json_format import VLLMLogitsProcessorJSON
from happy_vllm.logits_processors.response_pool import VLLMLogitsProcessorResponsePool

# In the following parser, we have entries of the form:
# api_keyword_main : {'class': a_logits_processor_class
#                'arguments': {api_keyword: argument_of_the_class,
#                              api_keyword: argument_of_the_class}, ...},
#                  'tokenizer': False, 'model': False, 'prompt': False
# The purpose of this parser is that, if the api_keyword_main is present in the json sent via API, then
# we know we must use the corresponding logits processor. Then we look for all keyword in the 'arguments' 
# dictionary which we each corresponds to an argument of the init of the class except for the tokenizer, the model or the prompt 
# which are passed by the tokenizer, model or prompt boolean values if needed

logits_processors_parser = {'response_pool': {'class': VLLMLogitsProcessorResponsePool,
                                                'arguments': {'response_pool': 'possible_responses'},
                                                'tokenizer': True},
                            'json_format':{'class': VLLMLogitsProcessorJSON,
                                            'arguments': {'json_format': 'json_format',
                                                          'json_format_is_json_schema': 'json_format_is_json_schema'},
                                            'tokenizer_lmformatenforcer': True}}

logits_processors_incompatibilities = [('response_pool', 'json_format')]

def detect_logits_processors_incompatibilities(request_dict: dict) -> None:
    """Detects if some keywords present in request_dict could allow incompatible logits_processors
    to be present

    Args:
        request_dict (dict) : The body of the request with the different keywords
    """
    for first_keyword, second_keyword in logits_processors_incompatibilities:
        if first_keyword in request_dict and second_keyword in request_dict:
            raise ValueError(f"Two keywords are incompatible : {first_keyword} and {second_keyword}")


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import ssl
import json

from typing import Optional
from pydantic_settings import BaseSettings, SettingsConfigDict
from argparse import Namespace, ArgumentParser, BooleanOptionalAction

from vllm.engine.arg_utils import AsyncEngineArgs


DEFAULT_MODEL_NAME = '?'
DEFAULT_APP_NAME = "happy_vllm"
DEFAULT_API_ENDPOINT_PREFIX = ""
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = 5000
DEFAULT_EXPLICIT_ERRORS = False
DEFAULT_ALLOW_CREDENTIALS = False
DEFAULT_ALLOWED_ORIGINS = ["*"]
DEFAULT_ALLOWED_METHODS = ["*"]
DEFAULT_ALLOWED_HEADERS = ["*"]
DEFAULT_UVICORN_LOG_LEVEL = 'info'
CHOICES_UVICORN_LOG_LEVEL = ['debug', 'info', 'warning', 'error', 'critical', 'trace']
DEFAULT_SSL_KEYFILE = None
DEFAULT_SSL_CERTFILE = None
DEFAULT_SSL_CA_CERTS = None
DEFAULT_SSL_CERT_REQS = int(ssl.CERT_NONE)
DEFAULT_ROOT_PATH = None


class ApplicationSettings(BaseSettings):
    """Application settings

    This class is used for settings management purpose, have a look at the pydantic
    documentation for more details : https://pydantic-docs.helpmanual.io/usage/settings/

    By default, it looks for environment variables (case insensitive) to set the settings
    if a variable is not found, it looks for a file name .env in your working directory
    where you can declare the values of the variables and finally it sets the values
    to the default ones one can define above
    """
    host : str = DEFAULT_HOST
    port: int = DEFAULT_PORT
    explicit_errors: bool = DEFAULT_EXPLICIT_ERRORS
    allow_credentials: bool = DEFAULT_ALLOW_CREDENTIALS
    allowed_origins: list = DEFAULT_ALLOWED_ORIGINS
    allowed_methods: list = DEFAULT_ALLOWED_METHODS
    allowed_headers: list = DEFAULT_ALLOWED_HEADERS
    uvicorn_log_level: str = DEFAULT_UVICORN_LOG_LEVEL
    ssl_keyfile: Optional[str] = DEFAULT_SSL_KEYFILE
    ssl_certfile: Optional[str] = DEFAULT_SSL_CERTFILE
    ssl_ca_certs: Optional[str] = DEFAULT_SSL_CA_CERTS
    ssl_cert_reqs: int = DEFAULT_SSL_CERT_REQS
    root_path: Optional[str] = DEFAULT_ROOT_PATH
    app_name: str = DEFAULT_APP_NAME
    api_endpoint_prefix: str = DEFAULT_API_ENDPOINT_PREFIX

    model_config = SettingsConfigDict(env_file=".env", extra='ignore', protected_namespaces=('settings', ))


def get_model_settings(parser: ArgumentParser) -> BaseSettings:
    """Gets the model settings. It corresponds to the variables added via AsyncEngineArgs.add_cli_args plus model-name.
    First we use the parser to get the default values of vLLM for these variables. We instantiate a BaseSettings model
    with these values as default. They are possibly overwritten by environnement variables or those of a .env

    Args:
        parser (ArgumentParser) : The parser containing all the model variables with thei default values from vLLM 
    """

    default_args = parser.parse_args([])

    # Define ModelSettings with the default values of the args (which will be replaced)
    # by the environnement variables or those of the .env file
    class ModelSettings(BaseSettings):
        model: str = default_args.model
        model_name: str = default_args.model_name
        tokenizer: Optional[str] = default_args.tokenizer
        tokenizer_mode: str = default_args.tokenizer_mode
        trust_remote_code: bool = False
        download_dir: Optional[str] = default_args.download_dir
        load_format: str = default_args.load_format
        dtype: str = default_args.dtype
        kv_cache_dtype: str = default_args.kv_cache_dtype
        seed: int = default_args.seed
        max_model_len: Optional[int] = default_args.max_model_len
        worker_use_ray: bool = False
        pipeline_parallel_size: int = default_args.pipeline_parallel_size
        tensor_parallel_size: int = default_args.tensor_parallel_size
        max_parallel_loading_workers: Optional[int] = default_args.max_parallel_loading_workers
        block_size: int = default_args.block_size
        enable_prefix_caching: bool = False
        swap_space: int = default_args.swap_space
        gpu_memory_utilization: float = default_args.gpu_memory_utilization
        max_num_batched_tokens: Optional[int] = default_args.max_num_batched_tokens
        max_num_seqs: int = default_args.max_num_seqs
        disable_log_stats: bool = False
        revision: Optional[str] = default_args.revision
        code_revision: Optional[str] = default_args.code_revision
        tokenizer_revision: Optional[str] = default_args.tokenizer_revision
        quantization: Optional[str] = default_args.quantization
        enforce_eager: bool = False
        max_context_len_to_capture: int = default_args.max_context_len_to_capture
        disable_custom_all_reduce: bool = False
        enable_lora: bool = False
        max_loras: int = default_args.max_loras
        max_lora_rank: int = default_args.max_lora_rank
        lora_extra_vocab_size: int = default_args.lora_extra_vocab_size
        lora_dtype: str = default_args.lora_dtype
        max_cpu_loras: Optional[int] = default_args.max_cpu_loras
        device: str = default_args.device
        ray_workers_use_nsight: bool = False
        max_log_len: Optional[int] = default_args.max_log_len
        disable_log_requests: bool = False
        engine_use_ray: bool = False
        use_v2_block_manager: bool = False
        num_lookahead_slots: int = default_args.num_lookahead_slots
        forced_num_gpu_blocks: Optional[int] = default_args.forced_num_gpu_blocks
        max_logprobs: int = default_args.max_logprobs
        tokenizer_pool_size: int = default_args.tokenizer_pool_size
        tokenizer_pool_type: str = default_args.tokenizer_pool_type
        tokenizer_pool_extra_config: Optional[str] = default_args.tokenizer_pool_extra_config
        image_input_type: Optional[str] = default_args.image_input_type
        image_input_shape: Optional[str] = default_args.image_input_shape
        image_feature_size: Optional[int] = default_args.image_feature_size
        scheduler_delay_factor: float = default_args.scheduler_delay_factor
        enable_chunked_prefill: bool = default_args.enable_chunked_prefill

        model_config = SettingsConfigDict(env_file=".env", extra='ignore', protected_namespaces=('settings', ))

    model_settings = ModelSettings()

    return model_settings


def get_parser() -> ArgumentParser:
    """Gets the parser. The default values of all application variables (see ApplicationSettings) are properly
    set to the BaseSetting value defined via pydantic. The default values of all model variables (ie those added
    via AsyncEngineArgs.add_cli_args plus model-name) are not properly set via pydantic at this point.
    
    Returns:
        ArgumentParser : The argparse parser
    """
    parser = ArgumentParser(description="REST API server for vLLM, production ready")

    application_settings = ApplicationSettings(_env_parse_none_str='None') # type: ignore

    parser.add_argument("--host",
                        type=str,
                        default=application_settings.host,
                        help="host name")
    parser.add_argument("--port",
                        type=int,
                        default=application_settings.port,
                        help="port number")
    parser.add_argument("--model-name",
                        type=str,
                        default=DEFAULT_MODEL_NAME,
                        help="The name of the model given by the /info endpoint of the API")
    parser.add_argument("--app-name",
                        type=str,
                        default=application_settings.app_name,
                        help="The name of the application")
    parser.add_argument("--api-endpoint-prefix",
                        type=str,
                        default=application_settings.api_endpoint_prefix,
                        help="The prefix for the API endpoints")
    parser.add_argument("--explicit-errors",
                        default=application_settings.explicit_errors,
                        action=BooleanOptionalAction,
                        help="If True, the underlying python errors are sent back via the API")
    parser.add_argument('--allow-credentials',
                        default=application_settings.allow_credentials,
                        action=BooleanOptionalAction,
                        help="allow credentials for CORS")
    parser.add_argument("--allowed-origins",
                        type=json.loads,
                        default=application_settings.allowed_origins,
                        help="CORS allowed origins")
    parser.add_argument("--allowed-methods",
                        type=json.loads,
                        default=application_settings.allowed_methods,
                        help="CORS allowed methods")
    parser.add_argument("--allowed-headers",
                        type=json.loads,
                        default=application_settings.allowed_headers,
                        help="CORS allowed headers")
    parser.add_argument("--uvicorn-log-level",
                        type=str,
                        default=application_settings.uvicorn_log_level,
                        choices=CHOICES_UVICORN_LOG_LEVEL,
                        help="log level for uvicorn")
    parser.add_argument("--ssl-keyfile",
                        type=str,
                        default=application_settings.ssl_keyfile,
                        help="The file path to the SSL key file")
    parser.add_argument("--ssl-certfile",
                        type=str,
                        default=application_settings.ssl_certfile,
                        help="The file path to the SSL cert file")
    parser.add_argument("--ssl-ca-certs",
                        type=str,
                        default=application_settings.ssl_ca_certs,
                        help="The CA certificates file")
    parser.add_argument("--ssl-cert-reqs",
                        type=int,
                        default=application_settings.ssl_cert_reqs,
                        help="Whether client certificate is required (see stdlib ssl module's)")
    parser.add_argument("--root-path",
                        type=str,
                        default=application_settings.root_path,
                        help="FastAPI root_path when app is behind a path based routing proxy")
    parser = AsyncEngineArgs.add_cli_args(parser)

    return parser


def parse_args() -> Namespace:
    """Parses the args. We want this priority : args from cli > environnement variables > .env variables.
    In order to do this, we get from pydantic BaseSetting the value from environnement variables or .env variables
    with the proper priority. Then we set the default value of the cli parser to those value so that if the cli args
    are used, they overwrite the default values and otherwise, the BaseSetting value is taken.

    Returns:
        NameSpace
    """
    # Gets the parser
    # The default value of the application variables are properly set
    # Those of the model variables are not
    parser = get_parser()
    # Gets the default values of the model variables via pydantic
    model_settings = get_model_settings(parser)
    # Sets the default values of the model variables in the parser
    parser.set_defaults(**model_settings.model_dump())
    # Gets the args
    args = parser.parse_args()

    # Explicitly check for help flag for the providing help message to the entrypoint
    if '-h' in sys.argv[1:] or '--help' in sys.argv[1:]:
        parser.print_help()
        sys.exit()
    return args



#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest
from happy_vllm.logits_processors.utils_parse_logits_processors import detect_logits_processors_incompatibilities


def test_detect_logits_processors_incompatibilities():
    request_dict = {"prompt": "my_prompt", "response_pool": ["Yes", "No"]}
    assert detect_logits_processors_incompatibilities(request_dict) is None

    request_dict = {"prompt": "my_prompt", "json_format": {"name": "Smith"}}
    assert detect_logits_processors_incompatibilities(request_dict) is None

    request_dict = {"prompt": "my_prompt", "response_pool": ["Yes", "No"], "json_format": {"name": "Smith"}}
    with pytest.raises(ValueError):
        detect_logits_processors_incompatibilities(request_dict)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Config global settings

This module handle global app configuration
"""


from pydantic_settings import BaseSettings, SettingsConfigDict


class LogSettings(BaseSettings):
    log_level: str = "INFO"

    model_config = SettingsConfigDict(env_file=".env", extra='ignore', protected_namespaces=('settings', ))


log_settings = LogSettings()


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from happy_vllm import utils_args

def test_get_parser():
    parser = utils_args.get_parser()
    application_settings = utils_args.ApplicationSettings()
    for key, value in application_settings.model_dump().items():
        assert parser.get_default(key) == value

def test_get_model_settings():
    parser = utils_args.get_parser()
    model_settings = utils_args.get_model_settings(parser)
    for key, field in model_settings.model_fields.items():
        if parser.get_default(key) is not None:
            assert parser.get_default(key) == field.default
            
    assert model_settings.model == os.environ['MODEL']
    assert model_settings.model_name == os.environ['MODEL_NAME']




#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import logging

from .core.config import log_settings
from .core.logtools import get_pattern_log

level = log_settings.log_level

logger = logging.getLogger(__name__)
logger.setLevel(level)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatterJson = logging.Formatter(get_pattern_log())
ch.setFormatter(formatterJson)

logger.addHandler(ch)

# tensorflow
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""This module contains the base Model class
"""


import os
import logging
from pathlib import Path
from argparse import Namespace
from transformers import AutoTokenizer
from typing import Any, Tuple, Union, List
from vllm.engine.arg_utils import AsyncEngineArgs
from vllm.engine.async_llm_engine import AsyncLLMEngine
from vllm.transformers_utils.tokenizer_group.tokenizer_group import TokenizerGroup
from lmformatenforcer.integrations.transformers import build_token_enforcer_tokenizer_data

from happy_vllm import utils

logger = logging.getLogger(__name__)


class Model:
    """Parent model class.
    """

    def __init__(self, **kwargs):
        '''Init. model class'''
        self._model = None
        self._tokenizer = None
        self._model_conf = None
        self._model_explainer = None
        self._loaded = False
        self.app_name = kwargs.get('app_name', "happy_vllm")

    def is_model_loaded(self):
        """return the state of the model"""
        return self._loaded

    def loading(self, args: Namespace, **kwargs):
        """load the model"""
        self._load_model(args, **kwargs)
        self._loaded = True

    def _load_model(self, args: Namespace, **kwargs) -> None:
        """Load a model from a file

        Returns:
            Tuple[Any, dict]: A tuple containing the model and a dict of metadata about it.
        """

        self._model_conf = {'model_name': args.model_name}

        logger.info(f"Loading the model from {args.model}")
        if args.model_name != "TEST MODEL":
            engine_args = AsyncEngineArgs.from_cli_args(args) 
            self._model = AsyncLLMEngine.from_engine_args(engine_args) # type: ignore
            if isinstance(self._model.engine.tokenizer, TokenizerGroup): # type: ignore
                self._tokenizer = self._model.engine.tokenizer.tokenizer # type: ignore
            else:
                self._tokenizer = self._model.engine.tokenizer # type: ignore
            self._tokenizer_lmformatenforcer = build_token_enforcer_tokenizer_data(self._tokenizer)
            self.max_model_len = self._model.engine.model_config.max_model_len # type: ignore
            self.original_truncation_side = self._tokenizer.truncation_side
        # For test purpose
        else:
            self.max_model_len = 2048
            self.original_truncation_side = 'right'
            self._tokenizer = AutoTokenizer.from_pretrained(utils.TEST_TOKENIZER_NAME,
                                                     cache_dir=os.environ["TEST_MODELS_DIR"], truncation_side=self.original_truncation_side)
            self._tokenizer_lmformatenforcer = build_token_enforcer_tokenizer_data(self._tokenizer)
            self._model = MockModel(self._tokenizer, self.app_name)
        logger.info(f"Model loaded")

    def tokenize(self, text: str) -> List[int]:
        """Tokenizes a text
        
        Args:
            text (str) : The text to tokenize

        Returns:
            list : The list of token ids
        """
        return list(utils.proper_tokenization(self._tokenizer, text))

    def split_text(self, text: str, num_tokens_in_chunk: int = 200, separators: Union[list, None] = None) -> List[str]:
        '''Splits a text in small texts containing at least num_tokens_in_chunk tokens and ending by a separator. note that the `separators` 
        used are the tokenization of the strings and not the strings themselves (which explains why we must for example 
        specify ' .' and '.' as two separate separators) 

        Args:
            text (str) : The text to split

        Kwargs:
            num_tokens_in_chunk (int) : The minimal number of tokens in the chunk
            separators (list) : The separators marking the end of a sentence

        Returns:
            A list of strings each string containing at least num_tokens_in_chunk tokens and ending by a separator
        '''
        if separators is None:
            separators = [".", "!", "?", "|", " .", " !", " ?", " |"]
        separators_tokens_ids = set()
        for separator in separators:
            separators_tokens_ids.add(utils.proper_tokenization(self._tokenizer, separator))
        tokens = list(utils.proper_tokenization(self._tokenizer, text))
        indices_separators = []
        for separator_tokens_ids in separators_tokens_ids:
            indices_separators += find_indices_sub_list_in_list(tokens, list(separator_tokens_ids))
        indices_separators.sort()

        chunks = []
        index_beginning_chunk = 0
        current_used_separator = 0
        while current_used_separator < len(indices_separators):
            index_current_used_separator = indices_separators[current_used_separator]
            if index_current_used_separator +1 - index_beginning_chunk >= num_tokens_in_chunk:
                chunks.append(tokens[index_beginning_chunk:index_current_used_separator + 1])
                index_beginning_chunk = index_current_used_separator + 1
            current_used_separator += 1
        chunks.append(tokens[index_beginning_chunk:])
        chunks = [utils.proper_decode(self._tokenizer, chunk) for chunk in chunks]
        chunks = [element for element in chunks if element!= ""]
        return chunks

    def extract_text_outside_truncation(self, text: str, truncation_side: Union[str, None] = None, max_length: Union[int, None] = None) -> str:
        """Extracts the part of the prompt not kept after truncation, which will not be infered by the model.
        First, we tokenize the prompt while applying truncation.
        We obtain a list of sequences of token ids padded, which are outside the truncation.
        Then we decode this list of tensors of token IDs containing special tokens to a string.

        Args:
            text (str) : The text we want to parse
            truncation_side (str) : The side of the truncation
            max_length (int) : The length above which the text will be truncated

        Returns:
            The part of the text which will be dropped by the truncation (str)
        """
        if max_length is None:
            max_length = self.max_model_len
        if truncation_side is None:
            truncation_side = self.original_truncation_side
        self._tokenizer.truncation_side = truncation_side
        list_tokens = self._tokenizer(text, truncation=True, add_special_tokens=False, max_length=max_length, return_overflowing_tokens=True)['input_ids']
        if len(list_tokens) <= 1:
            return ''
        not_truncated = list_tokens[0]
        truncated_tmp = list_tokens[1:]
        if self._tokenizer.truncation_side == 'left':
            truncated_tmp.reverse()
        truncated = []
        for truncated_tokens in truncated_tmp:
            truncated += truncated_tokens
        truncated_str = self._tokenizer.decode(truncated)
        self._tokenizer.truncation_side = self.original_truncation_side
        return truncated_str

    def get_gpu_kv_cache_usage(self) -> float:
        """Gets the GPU KV cache usage

        Returns:
            The GPU KV cache usage
        """
        total_num_gpu_blocks = self._model.engine.cache_config.num_gpu_blocks
        num_free_gpu_blocks = (
            self._model.engine.scheduler.block_manager.get_num_free_gpu_blocks())
        num_used_gpu_blocks = total_num_gpu_blocks - num_free_gpu_blocks
        gpu_cache_usage = num_used_gpu_blocks / total_num_gpu_blocks
        return gpu_cache_usage

    def get_cpu_kv_cache_usage(self) -> float:
        """Gets the CPU KV cache usage

        Returns:
            The CPU KV cache usage
        """
        total_num_cpu_blocks = self._model.engine.cache_config.num_cpu_blocks
        if total_num_cpu_blocks > 0:
            num_free_cpu_blocks = (
                self._model.engine.scheduler.block_manager.get_num_free_cpu_blocks())
            num_used_cpu_blocks = total_num_cpu_blocks - num_free_cpu_blocks
            cpu_cache_usage = num_used_cpu_blocks / total_num_cpu_blocks
        else:
            cpu_cache_usage = 0.0
        return cpu_cache_usage

    def get_status_requests(self) -> dict:
        """Gets the status of the different requests being processed

        Returns:
            A dictionary containing the number of requests in the different status (running, swapped and pending)
        """
        status_requests = {"requests_running": len(self._model.engine.scheduler.running),
                            "requests_swapped": len(self._model.engine.scheduler.swapped),
                            "requests_pending": len(self._model.engine.scheduler.waiting)}
        return status_requests


def find_indices_sub_list_in_list(big_list: list, sub_list: list) -> list:
    """Find the indices of the presence of a sub list in a bigger list. For example
    if big_list = [3, 4, 1, 2, 3, 4, 5, 6, 3, 4] and sub_list = [3, 4],
    the result will be [1, 5, 9]

    Args:
        big_list (list) : The list in which we want to find the sub_list
        sub_list (list): The list we want the indices of in the big_list

    Returns:
        list : The list of indices of where the sub_list is in the big_list 
    """
    len_sub_list = len(sub_list)
    indices = []
    for index in range(len(big_list)):
        if big_list[index - len_sub_list + 1: index + 1] == sub_list:
            indices.append(index)
    return indices


class MockModel():

    def __init__(self, tokenizer, app_name: str = "happy_vllm"):
        self.tokenizer = tokenizer
        self.app_name = app_name

    async def generate(self, prompt, sampling_params, request_id):
        stream_txts = [f"n={i} "*i + prompt + " This is the generated text. I find it really good don't you ?" for i in range(sampling_params.n)]
        stream_ids = [self.tokenizer(stream_txt, add_special_tokens=False, truncation=True, max_length=sampling_params.max_tokens)['input_ids'] for stream_txt in stream_txts]
        max_length = max([len(element) for element in stream_ids]) + 1
        stream_tmp = [[self.tokenizer.decode(text[:i]) for text in stream_ids] for i in range(max_length)]
        stream = [MockGenerateResponse(prompt, texts) for texts in stream_tmp]
        # Mock the length finish_reason
        for i in range(sampling_params.n):
            if stream[-1].outputs[i].finish_reason is None:
                stream[-1].outputs[i].finish_reason = "length"
        stream = self.async_iter(stream)
        async for outputs in stream:
            yield outputs

    async def async_iter(self, my_list):
        for element in my_list:
            yield element


class MockGenerateResponse():

    def __init__(self, prompt, outputs):
        self.prompt = prompt
        outputs = [MockOutput(output) for output in outputs]
        self.outputs = outputs


class MockOutput():
    
    def __init__(self, text):
        self.text = text
        if text[-len("don't you ?"):] == "don't you ?":
            self.finish_reason = "stop"
        else:
            self.finish_reason = None


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastapi import Request
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware

# Class for returning error messages as the return of an failed (error 500) API call
class ExceptionHandlerMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        try:
            return await call_next(request)
        except Exception as e:
            return JSONResponse(
                status_code=500, 
                content={
                    'error': e.__class__.__name__, 
                    'messages': e.args
                }
            )


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Logs utilities

This module is used to define log pattern, add log filters, etc.
"""


from happy_vllm import utils


def get_pattern_log():
    app_version = utils.get_package_version()
    return (
        "{"
        '"date": "%(asctime)s", '
        '"level": "%(levelname)s", '
        '"message": "%(message)s", '
        f'"version": "{app_version}", '
        '"function": "File %(pathname)s, line %(lineno)d, in %(funcName)s", '
        '"logger": "%(name)s"'
        "}"
    )


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch
from typing import List, Union
from lmformatenforcer.integrations.vllm import VLLMLogitsProcessor
from lmformatenforcer import TokenEnforcerTokenizerData, JsonSchemaParser, TokenEnforcer


def check_simple_json(json_format: dict) -> bool:
    """Checks if the simple_json given is of the right form : a flat dictionary where the values are in the list ['string', 'integer', 'boolean', 'number']
    or is a list with a single element in it belonging to the list ['string', 'integer', 'boolean', 'number'] and the keywords are strings.
    So a dictionary of the following type:
    {"first_keyword": "string",
    "second_keyword": "integer",
    "third_keyword": "boolean",
    "fourth_keyword": "number",
    "fifth_keyword": ["string"]}

    Args:
        json_format (dict) : The dictionary containing the key and value we want to check
    """
    for key, value in json_format.items():
        if not isinstance(key, str):
            raise ValueError(f'The keyword {key} in the json schema is not a string')
        possible_values = ['string', 'integer', 'boolean', 'number']
        if isinstance(value, list):
            if len(value) != 1:
                raise ValueError(f'The value {value} corresponding to the keyword {key} appears to be a list not composed of a single element in {possible_values}')
            if value[0] not in possible_values:
                raise ValueError(f'The value {value} corresponding to the keyword {key} is a list where the single element in not in {possible_values}')
        elif value not in possible_values:
            raise ValueError(f'The value {value} corresponding to the keyword {key} is not in {possible_values}')
    return True


def get_json_format_parser(json_format: dict, is_json_schema: bool) -> JsonSchemaParser:
    """Get the JsonSchemaParser from the json format indicated. This json format can be of a simple form as described in check_simple_json
    or is a valid json schema as described in https://json-schema.org/learn/getting-started-step-by-step

    Args:
        json_format (dict) : The json format we want
        is_json_schema (bool) : Indicates if the json_format is a valid json schema or not

    Returns:
        The lmformatenforcer JsonSchemaParser

    """
    if is_json_schema:
        schema = json_format
    else:
        # Check if json_format is indeed a json in a simple form
        check_simple_json(json_format)
        schema = {'properties':{},
                'required':[],
                'title':'simple_json',
                'type': 'object'}
        for key, type_value in json_format.items():
            if isinstance(type_value, list):
                schema['properties'][key] = {'type': 'array', 'items': {"type": type_value[0]}}
            else:
                schema['properties'][key] = {'type': type_value}
            schema['required'].append(key)

    json_format_parser = JsonSchemaParser(schema)
    return json_format_parser


class VLLMLogitsProcessorJSON(VLLMLogitsProcessor):

    def __init__(self, tokenizer_lmformatenforcer: TokenEnforcerTokenizerData, json_format: Union[dict, None], json_format_is_json_schema: bool = False) -> None:
        '''Initializes the logits processor for json outputs

        Args:
            tokenizer_lmformatenforcer : The tokenizer obtained via lm-format-enforcer
            json_format : The json format we want

        Kwargs:
            json_format_is_json_schema (bool) : Indicates if the json_format is a valid json schema or not
        '''
        # Tests if the input is relevant and thus if this logits processor should do nothing
        if json_format == {} or json_format is None:
            self.logits_processor_activated = False
        else:
            self.logits_processor_activated = True
            json_format_parser = get_json_format_parser(json_format, json_format_is_json_schema)
            token_enforcer = TokenEnforcer(tokenizer_lmformatenforcer, json_format_parser)
            super().__init__(token_enforcer, False)
    
    def __call__(self, input_ids: List[int], scores: torch.Tensor):
        if self.logits_processor_activated :
            return super(VLLMLogitsProcessorJSON, self).__call__(input_ids, scores)
        else:
            return scores
            


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import uvicorn
import argparse
from vllm.engine.arg_utils import AsyncEngineArgs

from happy_vllm.utils_args import parse_args
from happy_vllm.application import declare_application



def main() -> None:
    args = parse_args()

    app = declare_application(args=args)
    uvicorn.run(app,
                host=args.host,
                port=args.port,
                log_level=args.uvicorn_log_level,
                ssl_keyfile=args.ssl_keyfile,
                ssl_certfile=args.ssl_certfile,
                ssl_ca_certs=args.ssl_ca_certs,
                ssl_cert_reqs=args.ssl_cert_reqs)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import json
from vllm.utils import random_uuid
from fastapi import APIRouter, Body
from pydantic import BaseModel, Field
from starlette.requests import Request
from vllm.sampling_params import SamplingParams
from vllm.engine.async_llm_engine import AsyncLLMEngine
from lmformatenforcer import TokenEnforcerTokenizerData
from typing import Annotated, AsyncGenerator, Tuple, List
from transformers.tokenization_utils_base import PreTrainedTokenizerBase
from starlette.responses import JSONResponse, Response, StreamingResponse

from happy_vllm import utils
from happy_vllm.logits_processors.response_pool import VLLMLogitsProcessorResponsePool
from happy_vllm.logits_processors.utils_parse_logits_processors import logits_processors_parser, detect_logits_processors_incompatibilities

from ..model.model_base import Model
from ..core.resources import RESOURCE_MODEL, RESOURCES
from .schemas.functional import ResponseGenerate, RequestGenerate, ResponseTokenizer, RequestTokenizer, ResponseDecode, RequestDecode, ResponseSplitText, RequestSplitText, ResponseMetadata, RequestMetadata


# Load the response examples
directory = os.path.dirname(os.path.abspath(__file__))
request_examples_path = os.path.join(directory, "schemas", "examples", "request.json")
with open(request_examples_path, 'r') as file:
    request_openapi_examples = json.load(file)

# Functional router
router = APIRouter()


def parse_logits_processors(request_dict: dict, prompt: str, model: AsyncLLMEngine, tokenizer: PreTrainedTokenizerBase,
                            tokenizer_lmformatenforcer: TokenEnforcerTokenizerData) -> list:
    """Parses the body of the request in order to provide the logits processors
    
    Args:
        request_dict (dict): The body of the request
        model (AsyncLLMEngine): The model
        tokenizer : The tokenizer
        tokenizer_lmformatenforcer : The LM format enforcer version of the tokenizer

    Returns:
        list : The list of logits processors
    """
    logits_processors = []
    references = {'prompt': prompt, 'model': model, 'tokenizer': tokenizer, 'tokenizer_lmformatenforcer': tokenizer_lmformatenforcer}
    list_keyword = list(request_dict)
    for keyword_main in list_keyword:
        # Parse the body sent via API as explained in happy_vllm.logits_processors.utils_parse_logits_processors
        if keyword_main in logits_processors_parser:
            dict_logits_processor = logits_processors_parser[keyword_main]
            class_to_instantiate = dict_logits_processor['class']
            kwargs = {}     
            # Add prompt, model or tokenizer if needed
            for reference, refered_object in references.items():
                if dict_logits_processor.get(reference, False):
                    kwargs[reference] = refered_object
            # Add arguments to instantiate the logits_processor if needed
            for keyword, class_argument in dict_logits_processor['arguments'].items():
                if keyword in request_dict:
                    kwargs[class_argument] = request_dict.pop(keyword)
            logit_processor = class_to_instantiate(**kwargs)
            logits_processors.append(logit_processor)
    # To ensure that no unsuitable keywords are passed to vllm, we pop them
    for keyword_main, logits_processor_config in logits_processors_parser.items():
        for keyword in logits_processor_config['arguments']:
            if keyword in request_dict:
                request_dict.pop(keyword)
    return logits_processors


def parse_generate_parameters(request_dict: dict, model: AsyncLLMEngine, tokenizer: PreTrainedTokenizerBase,
                            tokenizer_lmformatenforcer: TokenEnforcerTokenizerData) -> Tuple[str, bool, SamplingParams]:
    """Parses the body of the request to obtain the prompt and the sampling parameters

    Args:
        request_dict (dict): The body of the request
        model (AsyncLLMEngine): The model
        tokenizer : The tokenizer
        tokenizer_lmformatenforcer : The LM format enforcer version of the tokenizer
    Returns:
        str : The prompt
        bool : Whether the prompt should be displayed in the response
        SamplingParams : The vllm sampling parameters
    """
    prompt = request_dict.pop("prompt")
    if 'prompt_in_response' in request_dict:
        prompt_in_response = request_dict.pop('prompt_in_response')
    else:
        prompt_in_response = False
    detect_logits_processors_incompatibilities(request_dict)
    logits_processors = parse_logits_processors(request_dict, prompt, model, tokenizer, tokenizer_lmformatenforcer)
    sampling_params = SamplingParams(**request_dict)
    sampling_params.logits_processors = logits_processors
    for logits_processor in logits_processors:
        if isinstance(logits_processor, VLLMLogitsProcessorResponsePool):
            min_tokens = sampling_params.min_tokens
            reponse_pool = logits_processor.possible_tokens_responses
            if min_tokens > 0 and len(reponse_pool):
                raise ValueError(f"min_tokens : {min_tokens} is incompatible with the `response_pool` keyword")
    return prompt, prompt_in_response, sampling_params


@router.post("/generate", response_model=ResponseGenerate)
async def generate(
    request: Request,
    request_type: Annotated[
        RequestGenerate,
        Body(openapi_examples=request_openapi_examples["generate"])] = None
    ) -> Response:
    """Generate completion for the request.

    The request should be a JSON object with the following fields:
    - prompt: The prompt to use for the generation.
    - other fields: The sampling parameters (See `SamplingParams` from vLLM for more details : 
                                             https://github.com/vllm-project/vllm/blob/main/vllm/sampling_params.py).
    """
    
    model: Model = RESOURCES.get(RESOURCE_MODEL)
    request_dict = await request.json()
    prompt, prompt_in_response, sampling_params = parse_generate_parameters(request_dict, model._model, model._tokenizer, model._tokenizer_lmformatenforcer)
    request_id = random_uuid()
    model._tokenizer.truncation_side = model.original_truncation_side
    results_generator = model._model.generate(prompt, sampling_params, request_id)

    # Non-streaming case
    final_output = None
    async for request_output in results_generator:
        if await request.is_disconnected():
            # Abort the request if the client disconnects.
            await model._model.abort(request_id)
            return Response(status_code=499)
        final_output = request_output

    if final_output is None:
        raise ValueError('The final ouput is None')
    prompt = final_output.prompt
    text_outputs = [output.text for output in final_output.outputs]
    finish_reasons = [output.finish_reason for output in request_output.outputs]
    finish_reasons = ["None" if finish_reason is None else finish_reason for finish_reason in finish_reasons]
    ret = {"responses": text_outputs, "finish_reasons": finish_reasons}
    if prompt_in_response:
        ret['prompt'] = prompt
    return JSONResponse(ret)


@router.post("/generate_stream", response_model=ResponseGenerate)
async def generate_stream(request: Request,
    request_type: Annotated[
        RequestGenerate,
        Body(openapi_examples=request_openapi_examples["generate_stream"])] = None
    ) -> StreamingResponse:
    """Generate completion for the request.

    The request should be a JSON object with the following fields:
    - prompt: The prompt to use for the generation.
    - other fields: The sampling parameters (See `SamplingParams` for details).
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)
    request_dict = await request.json()
    prompt, prompt_in_response, sampling_params = parse_generate_parameters(request_dict, model._model, model._tokenizer, model._tokenizer_lmformatenforcer)
    request_id = random_uuid()
    model._tokenizer.truncation_side = model.original_truncation_side
    results_generator = model._model.generate(prompt, sampling_params, request_id)

    async def stream_results() -> AsyncGenerator[str, None]:
        async for request_output in results_generator:
            prompt = request_output.prompt
            text_outputs = [
                output.text for output in request_output.outputs
            ]
            finish_reasons = [output.finish_reason for output in request_output.outputs]
            finish_reasons = ["None" if finish_reason is None else finish_reason for finish_reason in finish_reasons]
            ret = {"responses": text_outputs, "finish_reasons": finish_reasons}
            if prompt_in_response:
                ret['prompt'] = prompt
            yield (json.dumps(ret) + "\n")#.encode("utf-8")

    return StreamingResponse(stream_results())


@router.post("/tokenizer", response_model=ResponseTokenizer)
async def tokenizer(request: Request,
    request_type: Annotated[
        RequestTokenizer,
        Body(openapi_examples=request_openapi_examples["tokenizer"])] = None
    ) -> Response:
    """Tokenizes a text

    The request should be a JSON object with the following fields:
    - text: The text to tokenize
    - with_tokens_str (optional): Whether we want the tokens strings in the output
    - vanilla (optional) : Whether we want the vanilla version of the tokenizers
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)
    request_dict = await request.json()
    text = request_dict.pop("text")
    vanilla = request_dict.get("vanilla", True)
    with_tokens_str = request_dict.get('with_tokens_str', False)
    
    if vanilla:
        tokens_ids = model._tokenizer(text)['input_ids']
        if with_tokens_str:
            tokens_str = model._tokenizer.convert_ids_to_tokens(tokens_ids)
    else:
        tokens_ids = model.tokenize(text)
        if with_tokens_str:
            tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in tokens_ids]
            

    ret = {"tokens_ids": tokens_ids, "tokens_nb": len(tokens_ids)}
    if with_tokens_str:
        ret['tokens_str'] = tokens_str
    return JSONResponse(ret)


@router.post("/decode", response_model=ResponseDecode)
async def decode(request: Request,
    request_type: Annotated[
        RequestDecode,
        Body(openapi_examples=request_openapi_examples["decode"])] = None
    ) -> Response:
    """Decodes token ids

    The request should be a JSON object with the following fields:
    - token_ids: The ids of the tokens
    - with_tokens_str : If the result should also include a list of str
    - vanilla (optional) : Whether we want the vanilla version of the tokenizers
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)
    request_dict = await request.json()
    token_ids = request_dict.pop("token_ids")
    with_tokens_str = request_dict.get("with_tokens_str", False)
    vanilla = request_dict.get("vanilla", True)
    
    if vanilla:
        decoded_string = model._tokenizer.decode(token_ids)
        if with_tokens_str:
            tokens_str = model._tokenizer.convert_ids_to_tokens(token_ids)
    else:
        decoded_string = utils.proper_decode(model._tokenizer, token_ids)
        if with_tokens_str:
            tokens_str = [utils.proper_decode(model._tokenizer, token_id) for token_id in token_ids]
            

    ret = {"decoded_string": decoded_string}
    if with_tokens_str:
        ret[ "tokens_str"] = tokens_str
    return JSONResponse(ret)


@router.post("/split_text", response_model=ResponseSplitText)
async def split_text(request: Request,
    request_type: Annotated[
        RequestSplitText,
        Body(openapi_examples=request_openapi_examples["split_text"])] = None
    ):
    """Splits a text with a minimal number of token in each chunk. Each chunk is delimited by a separator

    The request should be a JSON object with the following fields:
    - text: The text to split
    - num_tokens_in_chunk (optional): The minimal number of tokens we want in each chunk
    - separators (optional) : The allowed separators between the chunks
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)

    request_dict = await request.json()
    split_text = model.split_text(**request_dict)
    response = {"split_text": split_text}

    return JSONResponse(response)


@router.post("/metadata_text", response_model=ResponseMetadata)
async def metadata_text(request: Request,
    request_type: Annotated[
        RequestMetadata,
        Body(openapi_examples=request_openapi_examples["metadata_text"])] = None):
    """Gives meta data on a text

    The request should be a JSON object with the following fields:
    - text: The text to parse
    - truncation_side (optional): The truncation side of the tokenizer
    - max_length (optional) : The max length before truncation

    The default values for truncation_side and max_length are those of the underlying model
    """
    model: Model = RESOURCES.get(RESOURCE_MODEL)

    request_dict = await request.json()

    tokens_ids = model.tokenize(request_dict['text'])
    truncated_text = model.extract_text_outside_truncation(**request_dict)
    ret = {"tokens_nb": len(tokens_ids), "truncated_text": truncated_text}

    return JSONResponse(ret)


#!/usr/bin/env python3
# Copyright (C) <2018-2024>  <Agence Data Services, DSI France Travail>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functional schemas"""

import os
import json
from pydantic import BaseModel, Field
from starlette.responses import JSONResponse
from typing import Any, List, Union, Optional

from .utils import NumpyArrayEncoder

# Load the response examples
directory = os.path.dirname(os.path.abspath(__file__))
response_examples_path = os.path.join(directory, "examples", "response.json")
with open(response_examples_path, "r") as file:
    response_examples = json.load(file)


class NumpyJSONResponse(JSONResponse):
    def render(self, content: Any) -> bytes:
        return json.dumps(content, cls=NumpyArrayEncoder).encode()


class ResponseGenerate(BaseModel):
    responses: List[str] = Field(None, title="List of responses")
    finish_reasons: List[str] = Field(None, title="List of finish reasons")
    prompt: str = Field(None, title="Input prompt")
    model_config = {"json_schema_extra": {"examples": [response_examples["generate"]]}}


# See https://github.com/vllm-project/vllm/blob/0ce0539d4750f9ebcd9b19d7085ca3b934b9ec67/vllm/sampling_params.py
# for parameters description
class RequestGenerate(BaseModel):
    prompt: str = Field(None, title="Input prompt")
    n: int = Field(None)
    best_of: Optional[int] = Field(None)
    presence_penalty: float = Field(None)
    frequency_penalty: float = Field(None)
    repetition_penalty: float = Field(None)
    temperature: float = Field(None)
    top_p: float = Field(None)
    top_k: int = Field(None)
    min_p: float = Field(None)
    seed: Optional[int] = Field(None)
    use_beam_search: bool = Field(None)
    length_penalty: float = Field(None)
    early_stopping: Union[bool, str] = Field(None)
    stop: Optional[Union[str, List[str]]] = Field(None)
    stop_token_ids: Optional[List[int]] = Field(None)
    include_stop_str_in_output: bool = Field(None)
    ignore_eos: bool = Field(None)
    max_tokens: Optional[int] = Field(None)
    min_tokens: int = Field(None)
    logprobs: Optional[int] = Field(None)
    prompt_logprobs: Optional[int] = Field(None)
    detokenize: bool = Field(None)
    skip_special_tokens: bool = Field(None)
    spaces_between_special_tokens: bool = Field(None)
    truncate_prompt_tokens: int = Field(None)
    response_pool: list = Field(None)
    json_format: dict = Field(None)
    json_format_is_json_schema: bool = Field(None)


class ResponseTokenizer(BaseModel):
    tokens_ids: List[int] = Field(None, title="List of token ids")
    tokens_nb: int = Field(None, title="Number of tokens")
    tokens_str: List[str] = Field(None, title="List of decoded tokens")
    model_config = {"json_schema_extra": {"examples": [response_examples["tokenizer"]]}}


class RequestTokenizer(BaseModel):
    text: str = Field(None, title="Text to tokenize")
    with_token_str: bool = Field(None, title="Add strings of tokens")
    vanilla: bool = Field(None, title="Use vanilla tokenizer")


class ResponseDecode(BaseModel):
    decoded_string: str = Field(None, title="Decoded text")
    tokens_str: List[str] = Field(None, title="Decoded tokens")
    model_config = {"json_schema_extra": {"examples": [response_examples["decode"]]}}


class RequestDecode(BaseModel):
    token_ids: List[int] = Field(None, title="List of token ids")
    with_token_str: bool = Field(None, title="Add strings of tokens")
    vanilla: bool = Field(None, title="Use vanilla model")


class ResponseSplitText(BaseModel):
    split_text: List[str] = Field(None, title="Split text")
    model_config = {
        "json_schema_extra": {"examples": [response_examples["split_text"]]}
    }


class RequestSplitText(BaseModel):
    text: str = Field(None, title="Input text")
    num_tokens_in_chunk: int = Field(None, title="Min number of tokens per chunk")
    separators: List[str] = Field(None, title="List of separators")


class ResponseMetadata(BaseModel):
    tokens_nb: int = Field(None, title="Number of tokens")
    truncated_text: str = Field(None, title="Truncated text")
    model_config = {
        "json_schema_extra": {"examples": [response_examples["metadata_text"]]}
    }


class RequestMetadata(BaseModel):
    text: str = Field(None, title="Input text")
    truncation_side: str = Field(None, title="Side of truncation")
    max_length: int = Field(None, title="Max length before truncation")
